# TP Hachage



 
# Partie 1 - Réaliser une fonction de hachage minimale parfaite 

On travaillera sur `tp_2_miso_mphf.py`.

Tout d'abord, voyons ensemble l'idée de la construction de la MPHF.
Puis suivez les commentaires dans le code à compléter.
Vous devez 
- finaliser la fonction pour construire la MPHF (`construction_mphf`), 
- écrire la fonction pour obtenir le hash d'un élément par la MPHF (`get_hash_mphf`), 
- écrire la fonction pour créer une table de hachage avec cette mphf (`create_hash_table`)

Ecrivez vos réponses et commentaires dans ce document.

Puis décommenter `compare_taille` à la fin et expliquer les résultats.

    - Sur le graphique, nous observons que l'evolution de la taille de la table avec une mphf croît de manière linéaire, contrairement à celle d'un dictionnaire qui évolue par palliers de plus en plus élevés, jusqu'à atteindre un plateau. La taille finale est 5 fois plus importante pour le dictionnaire que pour la mphf. D'une part, cela s'explique par le fait que le dictionnaire doit conserver en mémoire les clefs, alors que la mphf propose un hachage individuel pour chaque clef et ne les sauvegarde pas. D'autre part, la mphf produit une table de taille correspondante au nombre d'éléments qu'on souhaite y enregistrer, contrairement au dictionnaire qui alloue progressivement de l'espace vide selon un doublage de la taille chaque fois que les deux tiers de l'espace sont pris.

Bonus : faites varier `nb_niveaux` et `gamma`, voyez quelle influence ils peuvent avoir.

`nb_niveaux`:
- Lorsqu'on augmente le nombre de niveaux :
    - réduit les collisions possibles car elles sont davantages réparties
    - peut augmenter la taille de la MPHF, car chaque niveau correspond à une nouvelle table de hachage
- Lorsqu'on diminue le nombre de niveaux :
    - le risque de collisions augmentes 
    - peut réduire la taille de la MPHF
Il faut donc un nombre de niveau suffisant pour éviter les collisions mais pas trop élevé afin de limiter l'espace utilisé par la MPHF.

# Partie 2 - Analyse de performance de dictionnaires en Python

On travaillera sur `tp_2_miso_dict.py`

A la fin de cette partie, en lançant `python tp_2_miso_dict.py` on doit obtenir 4 graphiques.

## Introduction

Nous allons analyser la performance de dictionnaires `dict` en Python en fonction de différents facteurs : que le facteur de charge (loading factor), 
le temps d'insertion et la taille de la mémoire occupée. Vous pourrez utiliser matplotlib et numpy pour visualiser les résultats.

##  Étude du facteur de charge

Fonction `experiment_load_factor`:

0. La fonction `experiment_load_factor` doit être définie pour prendre en entrée une liste de facteurs de charge (`load_factors`) et renvoyer les temps :
- d'insertion de clefs, 
- les nombres de réallocations de mémoire (quand le dictionnaire est ré-écrit en mémoire pour agrandir sa taille) 
- et les tailles de mémoire occupées par le dictionnaire pour chaque facteur de charge. 

1. Initialisation des listes pour stocker les résultats.
Initialiser dans la fonction `experiment_load_factor` des listes `insertion_times`, `num_resizes` et `sizes` seront utilisées pour stocker les trois résultats décrits au dessus.

2. La fonction `experiment_load_factor` doit boucler sur chaque facteur de charge de la liste en entrée `load_factors`. Créer un dictionnaire d'abord vide à chaque fois.

3. Initialiser les variables `num_elements`, `start_time`, `num_resize` et `last_size` pour respectivement mesurer  le nombre d'éléments, le temps d'insertion et le nombre de réallocations de mémoire.

4. Insérer les éléments dans le dictionnaire et mesurer le temps d'insertion pour chaque élément, vérifier le nombre de réallocations mémoire (utiliser par exemple `time.time()` et `sys.getsizeof()` pour mesurer le temps et la taille du dictionnaire avant et après chaque insertion).

La fonction time.time() a été remplacée par la fonction time.perf_counter() pour plus de précision.

5. Pour un facteur de charge donné, stocker les résultats dans les listes `insertion_times`, `num_resizes` et `sizes`.

## Deuxième étude

6. A quoi sert la fonction `experiment_longest` ?

cf Script 

## Visualisation des résultats

7. Créez quatre graphiques au format png ou pdf : 

1. - Un graphique du temps d'insertion en fonction du facteur de charge (obtenu question 5)
2. - Un graphique du nombre de réallocations de mémoire en fonction du facteur de charge (obtenu question 5)
3. - Un graphique de la taille de mémoire occupée en fonction du nombre d'éléments (obtenu question 5)
4. - Un histogramme des fréquences des temps d'insertions discrétisés (code fourni, remplacer la liste vide par la bonne entrée)

10. Commentez vos résultats.

1: On observe que plus les facteurs de charge sont importants, plus les temps d'insertion sont longs. En effet, on peut voir sur le graphique une courbe qui tend à croître. 
Ce phénomène peut être expliqué par les risques de collision accrus et le besoin de réallouer de la mémoire pour insérer un nouvel élément lorsque le facteur de charge est élevé. 

2: On a inséré 1 000 éléments dans le dictionnaire et on observe un nombre de réallocations de mémoire différent entre les facteurs de charge. 
Pour le facteur de charge le plus faible, 2 réallocations ont été réalisées contre 9 pour ceux allant de 0.7 à 1.0. On atteint alors un pallier où les risques de collisions sont assez faibles pour que Python n'est plus besoin de réallouer de la mémoire afin les éviter. On en déduit donc que pour les 1 000 éléments, après 9 réallocations, la mémoire ne dépasse pas les 2/3 de la taille du dictionnaire.
On observe également un plateau à 8 réallocations pour les facteurs 0.4, 0.5 et 0.6. On peut alors supposer que la réallocation supplémentaire pour les facteurs suivants est causée par le remplissage immédiat des 2/3 de la mémoire. En effet, Python réalise cette action lorsque le dictionnaire est rempli au 2/3 or, 2/3 = 0.66 ce qui pourrait expliquer le plateau suivant. 

3: Sur le graphique représentant la taille de la mémoire occupée en fonction du nombre d'éléments, que plus on a d'éléments dans le dictionnaire plus la taille de la mémoire occupée est importante. On retrouve une courbe semblable à celle correspondante au nombre de réallocations mémoire en fonction des facteurs de charge. Ceci est expliqué par le fait que selon le facteur de charge, de la mémoire est réalloué afin de permettre l'insertion d'éléments et que le facteur de charge correspond aux nombres d'éléments / taille interne de la table de hachage.

4: On peut voir sur l'histogramme que le temps d'insertion dans un dictionnaire Python est rapide. En effet, malgré les coûts liés à la mémoire lors de l'insertion, ils ont été optimisés afin de rendre efficace leur utilisation.