import matplotlib.pyplot as plt
import numpy as np
import time
import sys


###### PARTIE 2 ######

def mean (liste : list) -> float :
	"""
	"""
	taille = len(liste)
	return sum(liste)/taille

def experiment_load_factor(load_factors):
	"""
	Étude du facteur de charge
	"""
	insertion_times=[]
	num_resizes = []
	sizes = []

	for factor in load_factors :

		d = {}
		num_elements = int(factor*1000)
		count_resize = 0
		last_size = sys.getsizeof(d)
		start_time = time.perf_counter()
		
		for i in range(num_elements):
			d[i]=i
			current_size=sys.getsizeof(d)

			if current_size > last_size:
				count_resize += 1
				last_size = current_size

		stop_time = time.perf_counter()  # Temps d'insertion
		insertion_times.append(stop_time-start_time)
		
		num_resizes.append(count_resize)
		sizes.append(last_size)
		
	return insertion_times, num_resizes, sizes

def experiment_longest():
	"""
	Mesure le temps d'insertion de 10 000 paires clé-valeur dans un dictionnaire Python et 
	renvoie un histogramme des temps d'insertion sous forme d'un tableau.
	"""
	#Initalisation d'un dictionnaire
	d = {}
	#Liste qui contiendra le nombre d'insertions
	insertion_times = []

	#On parcourt de 0 à 9999
	for i in range(10000):
		#on définit une clé pour chaque valeur comme étant un int sous forme de str
		key = str(i)
		#on définit une variable valeur comme étant l'index
		value = i
		#on démarre un chrono
		start_time = time.time()
		#on ajoute la clé et sa valeur associée au dictionnaire
		d[key] = value
		#on mesure le temps d'insertion en stoppant le chrono et mesurant la différence entre le début et le stop
		insertion_time = time.time() - start_time
		#on l'ajoute à la liste des temps
		insertion_times.append(insertion_time)
	#on crée un histogramme à l'aide de numpy et on récupère la liste des temps d'insertions sous forme de tableau 
	frequencies = np.histogram(insertion_times)[0]
	#Renvoie les valeurs
	return frequencies

def visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies):
	"""
	Visualisation des résultats
	"""
	# Temps d'insertion en fonction du facteur de charge
	plt.figure(figsize=(10, 6))
	plt.plot(load_factors, insertion_times, marker = 'o', color="darkorange")
	plt.xlabel("Facteurs de charge")
	plt.ylabel("Temps d'execution")
	plt.xticks (load_factors)
	plt.title("Temps d'insertion en fonction du facteur de charge")
	plt.savefig("load_factors.png")
	
	# Nombre de réallocations de mémoire en fonction du facteur de charge
	plt.figure(figsize=(10, 6))
	plt.plot(load_factors, num_resizes, color="mediumvioletred", marker= 'o')
	plt.xlabel("Facteurs de charge")
	plt.ylabel("Nombre de réallocations")
	plt.xticks(load_factors)
	plt.title("Nombre de réallocations en fonction du facteur de charge")
	plt.savefig("num_resizes.png")

	# Taille de mémoire occupée en fonction du nombre d'éléments
	nombre_elements = [int(factor * 1000) for factor in load_factors]
	plt.figure(figsize=(10, 6))
	plt.plot(nombre_elements, sizes, color="darkturquoise", marker='o')
	plt.xlabel("Nombre d'éléments")
	plt.ylabel("Mémoire occupée")
	plt.xticks(nombre_elements)
	plt.title("Mémoire occupée en fonction du nombre d'éléments")
	plt.savefig("sizes.png")

	# Deuxième étude
	f = frequencies
	plt.figure(figsize=(10, 6))
	plt.bar(range(len(f)), f)
	plt.xlabel('Temps d\'insertion (s)')
	plt.ylabel('Fréquence')
	plt.title('Histogramme des fréquences des temps d\'insertions')
	plt.yscale('log')
	xticks = np.logspace(-6, 1, 3)  
	xtick_labels = [f'{x:.1e}' for x in xticks]  
	plt.xticks(xticks, xtick_labels)    
	plt.savefig('histogramme.png')

load_factors = [0.01, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
insertion_times, num_resizes, sizes = experiment_load_factor(load_factors)
frequencies = experiment_longest()
visualisation(load_factors, insertion_times, num_resizes, sizes, frequencies)
